from pathlib import Path
from typing import List, Set

from pytestarch.importer.import_types import Import


class ImporteeModuleCalculator:
    """Adds all parent modules of imported modules if they are not yet part of the modules list."""

    def __init__(self, root_path: Path) -> None:
        self._root_path = root_path

    def calculate_importee_modules(
        self,
        imports: List[Import],
        all_modules: List[str],
    ) -> List[str]:
        """For all imported modules: Calculate parent modules and add them to the list of existing modules if they
        are not already part of this list. This mainly applies to external dependencies.

        Args:
            imports:
            all_modules:

        Returns:
            all modules extended by parent modules of imported modules
        """
        extended_modules = set(all_modules)

        for imp in imports:
            importee = imp.importee()

            if str(self._root_path) not in importee:
                extended_modules.update(self._calculate_parent_modules(imp))

        return list(extended_modules)

    def _calculate_parent_modules(self, imp: Import) -> Set[str]:
        modules = {imp.importee()}
        modules.update(imp.importee_parent_modules())

        return modules



from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from typing import Set, List, Optional, Tuple

from pytestarch.eval_structure.eval_structure_types import Module, EvaluableArchitecture
from pytestarch.query_language.exceptions import ImproperlyConfigured


class ModuleRequirement:
    """Stores information about which module is supposed to be checked against which module."""

    def __init__(
        self, left_hand_module: Module, right_hand_module: Module, import_relation: bool
    ) -> None:
        self._left_hand_module = left_hand_module
        self._right_hand_module = right_hand_module

        self._right_hand_module_has_specifier = import_relation

        if self.left_hand_module_has_specifier:
            self._left_hand_module, self._right_hand_module = (
                self._right_hand_module,
                self._left_hand_module,
            )

    @property
    def left_hand_module(self) -> Module:
        return self._left_hand_module

    @property
    def right_hand_module(self) -> Module:
        return self._right_hand_module

    @property
    def left_hand_module_has_specifier(self) -> bool:
        return not self._right_hand_module_has_specifier


@dataclass
class RuleViolations:
    lax_dependencies_found: List[Module]
    strict_dependency: Optional[Tuple[str, str]]

    should_not_except_violated: bool = False
    should_only_except_violated_by_forbidden_import: bool = False
    should_not_violated: bool = False
    should_only_violated: bool = False
    should_only_except_violated_by_no_import: bool = False
    should_except_violated: bool = False
    should_violated: bool = False

    def __bool__(self) -> bool:
        return any(self.__dict__[a] for a in self._bool_types())

    @classmethod
    def _bool_types(cls) -> List[str]:
        return [field.name for field in fields(cls) if field.type == bool]


class BehaviorRequirement:
    """Stores information about which import behavior the checked module is supposed to exhibit."""

    def __init__(
        self,
        positive_behavior_lax: bool,
        positive_behavior_strict: bool,
        negative_behavior: bool,
        behavior_exception: bool,
    ) -> None:
        self._should = positive_behavior_lax
        self._should_only = positive_behavior_strict
        self._should_not = negative_behavior

        self._behavior_exception = behavior_exception

        self._validate()

    @property
    def strict_dependency_required(self) -> bool:
        """Returns True if a dependency between two specific modules is required."""
        return (self._should or self._should_only) and not self._behavior_exception

    @property
    def lax_dependency_required(self) -> bool:
        """Returns True if a dependency between a specific module and
        any other modules is required. Which other modules are required
        is specified by the module requirement."""
        return self._behavior_exception and (self._should or self._should_only)

    @property
    def strict_dependency_not_allowed(self) -> bool:
        """Returns True if a dependency between two specific modules is not allowed."""
        return (self._should_not and not self._behavior_exception) or (
            self._should_only and self._behavior_exception
        )

    @property
    def lax_dependency_not_allowed(self) -> bool:
        """Returns True if no dependency between a specific module
        and a set of other modules is not allowed. Which other modules
        are not allowed is specified by the module requirement."""
        return (self._should_not and self._behavior_exception) or (
            self._should_only and not self._behavior_exception
        )

    def _validate(self) -> None:
        if self.strict_dependency_required and self.strict_dependency_not_allowed:
            raise ImproperlyConfigured(
                "Strict dependency both required and not allowed."
            )

        if self.lax_dependency_required and self.lax_dependency_not_allowed:
            raise ImproperlyConfigured("Lax dependency both required and not allowed.")

    def generate_rule_violation(
        self,
        strict_dependency: Optional[Tuple[str, str]],
        lax_dependencies: Set[Module],
    ) -> RuleViolations:
        """Translate from the detected types of dependencies back to which behavior and dependency requirements are violated by them.
        Args:
            strict_dependency: dependency between the two specified modules found
            lax_dependencies: other dependencies found beside the two specified modules
        Returns:
            overview of all rule violations
        """
        should_not_except_expected = self._should_not and self._behavior_exception
        should_only_except_violated_by_forbidden_expected = (
            self._should_only and self._behavior_exception
        )
        should_not_expected = self._should_not and not self._behavior_exception
        should_only_expected = self._should_only and not self._behavior_exception
        should_only_except_violated_by_no_expected = (
            self._should_only and self._behavior_exception
        )
        should_except_expected = self._should and self._behavior_exception
        should_expected = self._should and not self._behavior_exception

        lax_dependencies_present = (
            lax_dependencies is not None and len(lax_dependencies) > 0
        )
        strict_dependency_present = strict_dependency is not None

        return RuleViolations(
            should_not_except_violated=should_not_except_expected
            and lax_dependencies_present,
            should_only_except_violated_by_forbidden_import=should_only_except_violated_by_forbidden_expected
            and (strict_dependency_present and lax_dependencies_present),
            should_not_violated=should_not_expected and strict_dependency_present,
            should_only_violated=should_only_expected
            and (not strict_dependency_present or lax_dependencies_present),
            should_only_except_violated_by_no_import=should_only_except_violated_by_no_expected
            and not lax_dependencies_present,
            should_except_violated=should_except_expected
            and not lax_dependencies_present,
            should_violated=should_expected and not strict_dependency_present,
            lax_dependencies_found=lax_dependencies,
            strict_dependency=strict_dependency,
        )


class RuleMatcher(ABC):
    """Checks whether given modules fulfill the module and behavior requirements that have been specified for them."""

    def __init__(
        self,
        module_requirement: ModuleRequirement,
        behavior_requirement: BehaviorRequirement,
    ) -> None:
        self._module_requirement = module_requirement
        self._behavior_requirement = behavior_requirement

    @abstractmethod
    def match(self, evaluable: EvaluableArchitecture) -> RuleViolations:
        """
        Checks whether an expected behavior is exhibited by the EvaluableArchitecture.

        Args:
            evaluable: object to check
        Returns:
            overview of rule violations
        """
        raise NotImplementedError()


class DefaultRuleMatcher(RuleMatcher):
    def match(self, evaluable: EvaluableArchitecture) -> RuleViolations:
        strict_dependency_required = (
            self._behavior_requirement.strict_dependency_required
        )
        lax_dependency_required = self._behavior_requirement.lax_dependency_required
        no_strict_dependency_allowed = (
            self._behavior_requirement.strict_dependency_not_allowed
        )
        no_lax_dependency_allowed = (
            self._behavior_requirement.lax_dependency_not_allowed
        )

        left_hand_module = self._module_requirement.left_hand_module
        right_hand_module = self._module_requirement.right_hand_module

        strict_dependency = None
        lax_dependencies = None

        if strict_dependency_required or no_strict_dependency_allowed:
            strict_dependency = evaluable.get_dependency(
                left_hand_module, right_hand_module
            )

        if lax_dependency_required or no_lax_dependency_allowed:
            if not self._module_requirement.left_hand_module_has_specifier:
                dependency_check_method = evaluable.any_dependency_to_module_other_than
            else:
                dependency_check_method = evaluable.any_other_dependency_to_module_than

            lax_dependencies = dependency_check_method(
                left_hand_module,
                right_hand_module,
            )

        return self._behavior_requirement.generate_rule_violation(
            strict_dependency, lax_dependencies
        )



import ast
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, List

from pytestarch.query_language.exceptions import ImproperlyConfigured


@dataclass
class NamedModule:
    """Contains an ast module with its name.

    Attributes:
        module: ast module object
        name: module name
    """

    module: ast.Module
    name: str


class Import(ABC):
    def __init__(self, importer: str) -> None:
        self._importer = importer

        self._importer_module_hierarchy = self._get_parent_modules(self._importer)

    def importer(self) -> str:
        """Returns name of the module that imports something.

        Returns:
            module name
        """
        return self._importer

    def importer_parent_modules(self) -> List[str]:
        """Returns names of all parent modules of the importing module.

        Returns:
            list of module names
        """
        return self._importer_module_hierarchy

    @abstractmethod
    def importee(self) -> str:
        """Returns name of a module that is being imported.

        Returns:
            module name
        """
        raise NotImplementedError()

    @abstractmethod
    def importee_parent_modules(self) -> List[str]:
        """Returns names of all parent modules of the imported module.

        Returns:
            list of module names
        """
        raise NotImplementedError()

    def __str__(self) -> str:
        return f"{self.importer()} imports {self.importee()}"

    def _get_parent_modules(self, module: str) -> List[str]:
        """Calculates all parent modules of a given module.

        Example: source root is a
        module: a.b.c
        returned: [a, a.b]

        Args:
            module: module to calculate parent modules for

        Returns:
            List of all parent modules, containing their full names up to the source code root.
        """
        parent_modules = []

        parent: List[str] = []

        for char in module:
            if char == ".":
                parent_modules.append("".join(parent))

            parent.append(char)

        return parent_modules


class AbsoluteImport(Import):
    """Represents an absolute import."""

    def __init__(self, importer: str, module_name: str) -> None:
        super().__init__(importer)
        self._module_name = module_name
        self._importee_module_hierarchy = self._get_parent_modules(self._module_name)

    def importee(self) -> str:
        return self._module_name

    def importee_parent_modules(self) -> List[str]:
        return self._importee_module_hierarchy


class RelativeImport(Import):
    """Represents a relative import."""

    def __init__(
        self,
        importer: str,
        module_name: Optional[str],
        import_name: Optional[str],
        level: int,
    ) -> None:
        super().__init__(importer)
        self._module_name = module_name
        self._import_name = import_name
        self._level = level

        if module_name is None and import_name is None:
            raise ImproperlyConfigured(
                "Either name of module of of import needs to be specified."
            )

        self._importee = self._calculate_importee()

        self._importee_module_hierarchy = self._get_parent_modules(
            self._module_name or self._import_name
        )

    def importee(self) -> str:
        return self._importee

    def importee_parent_modules(self) -> List[str]:
        return self._importee_module_hierarchy

    def _calculate_importee(self) -> str:
        module_name = self._module_name or self._import_name

        return self._importer_module_hierarchy[-self._level] + "." + module_name



"""
The following two functions are the main entry point to PyTestArch. They can be used to create an evaluable object,
for which the user can then define architectural rules.
"""
import os
from pathlib import Path
from types import ModuleType
from typing import Tuple, Optional

from pytestarch.config.config import Config
from pytestarch.eval_structure.eval_structure_types import EvaluableArchitecture
from pytestarch.eval_structure.evaluable_graph import EvaluableArchitectureGraph
from pytestarch.eval_structure.graph import Graph
from pytestarch.importer.converter import ImportConverter
from pytestarch.importer.file_filter import FileFilter
from pytestarch.importer.import_filter import ImportFilter
from pytestarch.importer.importee_module_calculator import ImporteeModuleCalculator
from pytestarch.importer.parser import Parser


DEFAULT_EXCLUSIONS = ("*__pycache__",)


def get_evaluable_architecture(
    root_path: str,
    module_path: str,
    exclusions: Tuple[str, ...] = DEFAULT_EXCLUSIONS,
    exclude_external_libraries: bool = True,
    level_limit: Optional[int] = None,
) -> EvaluableArchitecture:
    """Constructs an evaluable object based on the given module.

    Args:
        root_path: root directory of the source code
        module_path: path of module to generate the evaluable for. Must be a submodule of the root_path module.
        exclusions: pseudo-regex to exclude files and directories from being integrated into the evaluable, e.g. *Test.py
        exclude_external_libraries: if True, external dependencies (defined as all dependencies to module outside the
            module_path, including Python built-in modules) will not be taken into account for the dependency analysis.
        level_limit: if not None, specifies the depth of the graph. For example, a limit of 1 will result in only the
            modules one level below the module path to be added as nodes. Note that this only applies to the final graph;
            all modules will be parsed, the graph will simply be flattened: if a submodule of X imports Y, this import is
            then assigned to X instead, if Y is above the level limit.
    """
    root_path = Path(root_path)
    module_path = Path(module_path)

    root_and_module_path_identical = root_path == module_path

    if root_and_module_path_identical and level_limit is not None:
        # root directory will be appended to all paths, de facto increasing the depth by 1
        level_limit += 1

    config = Config(exclusions)
    file_filter = FileFilter(config)

    parser = Parser(root_path, file_filter, root_and_module_path_identical)
    converter = ImportConverter()

    all_modules, ast = parser.parse(module_path)

    imports = converter.convert(ast)

    import_filter = ImportFilter(exclude_external_libraries, module_path)
    imports = import_filter.filter(imports)

    if not exclude_external_libraries:
        all_modules = ImporteeModuleCalculator(root_path).calculate_importee_modules(
            imports,
            all_modules,
        )
    return EvaluableArchitectureGraph(Graph(all_modules, imports, level_limit))


def get_evaluable_architecture_for_module_objects(
    root_module: ModuleType,
    module: ModuleType,
    exclusions: Tuple[str, ...] = DEFAULT_EXCLUSIONS,
    exclude_external_libraries: bool = True,
    level_limit: Optional[int] = None,
) -> EvaluableArchitecture:
    """Same functionality as get_evaluable_architecture, but root module and module to evaluate are passed in as module objects
    instead of the absolute paths to them.
    """
    root_path = os.path.dirname(root_module.__file__)
    module_path = os.path.dirname(module.__file__)

    return get_evaluable_architecture(
        root_path, module_path, exclusions, exclude_external_libraries, level_limit
    )



import ast
from _ast import Import
from typing import Optional, List

from pytestarch.importer.import_types import NamedModule, AbsoluteImport, RelativeImport


class ImportConverter:
    """Converts all ast imports to custom import types."""

    def convert(self, asts: List[NamedModule]) -> List[Import]:
        """Converts ast modules to custom import modules. Filters out all modules
        that are not imports.

        Args:
            asts: list of ast modules
        Returns:
            list of import objects
        """
        module_to_search = asts
        imports = []

        while module_to_search:
            module = module_to_search.pop()

            ast_module, module_name = module.module, module.name

            if hasattr(ast_module, "body"):
                module_to_search.extend(
                    [NamedModule(m, module_name) for m in ast_module.body]
                )
            else:
                new_imports = self._convert(module.module, module.name)

                if new_imports:
                    imports.extend(new_imports)

        return imports

    def _convert(self, module: ast.Module, module_name: str) -> Optional[List[Import]]:
        """Calculates all imports of the given ast module.

        Args:
            module: module to convert
            module_name: name of the module

        Returns:
            list of calculated import objects
        """
        new_imports = None

        if isinstance(module, ast.Import):
            new_imports = []
            for alias in module.names:
                new_imports.append(AbsoluteImport(module_name, alias.name))
        elif isinstance(module, ast.ImportFrom):
            if module.level == 0:
                new_imports = [AbsoluteImport(module_name, module.module)]
            else:
                new_imports = []
                for alias in module.names:
                    new_imports.append(
                        RelativeImport(
                            module_name, module.module, alias.name, module.level
                        )
                    )

        return new_imports



"""Base class for different graph implementations of an evaluable structure. Delegates direct access to graph nodes
and edges to its subclasses in a template pattern.
"""
from typing import Set, Any, Optional, Tuple, List

from pytestarch.eval_structure.eval_structure_types import EvaluableArchitecture, Module
from pytestarch.eval_structure.graph import Graph, Node


class EvaluableArchitectureGraph(EvaluableArchitecture):
    """Abstract implementation of an evaluable object that is based on a graph structure."""

    def __init__(self, graph: Graph) -> None:
        self._graph = graph

    def get_dependency(
        self, dependent: Module, dependent_upon: Module
    ) -> Optional[Tuple[str, str]]:
        dependent_node = self._get_node(dependent)
        dependent_upon_nodes = self._get_all_submodules_of(dependent_upon)

        nodes_to_exclude = self._get_excluded_nodes([dependent, dependent_upon])

        nodes_to_check = [dependent_node]
        checked_nodes = set()

        while nodes_to_check:
            node = nodes_to_check.pop()

            if node in checked_nodes:
                continue

            checked_nodes.add(node)

            children = self._graph.direct_successor_nodes(node)

            for child in children:
                if self._graph.parent_child_relationship(node, child):
                    nodes_to_check.append(child)

                elif (
                    child in dependent_upon_nodes
                    and node not in nodes_to_exclude
                    and child not in nodes_to_exclude
                ):
                    return node, child

        return None

    def _get_node(self, dependent: Module) -> Node:
        return dependent.name or dependent.parent_module

    def any_dependency_to_module_other_than(
        self, dependent: Module, dependent_upon: Module
    ) -> List[Module]:
        # submodules of the dependent upon module do not count as an import that is not the dependent upon module
        nodes_to_exclude = self._get_all_submodules_of(dependent_upon)

        nodes_fulfilling_criteria = []
        # should submodules of the dependent module import each other, this does not count as a dependency
        nodes_that_do_not_fulfill_criterion = self._get_all_submodules_of(dependent)

        if dependent.parent_module is not None:
            # if the parent module is set and has a dependency other than dependent upon, it should not count as only
            # true submodules should be considered
            nodes_to_exclude.add(dependent.parent_module)

        if dependent_upon.parent_module is not None:
            # if there is a dependency to the parent module of dependent upon, this counts, as only dependencies to
            # the true submodules are excluded
            nodes_to_exclude.remove(dependent_upon.parent_module)

        nodes_to_check = list(nodes_that_do_not_fulfill_criterion)
        checked_nodes = set()

        while nodes_to_check:
            if nodes_fulfilling_criteria:
                break

            node = nodes_to_check.pop()

            if node in checked_nodes:
                continue

            if node in nodes_to_exclude:
                continue

            checked_nodes.add(node)

            children = self._graph.direct_successor_nodes(node)

            for child in children:

                if not self._graph.parent_child_relationship(node, child):
                    if (
                        child not in nodes_to_exclude
                        and child not in nodes_that_do_not_fulfill_criterion
                    ):
                        nodes_fulfilling_criteria.append(child)

                    nodes_to_check.append(child)

        return self._to_modules(nodes_fulfilling_criteria)

    def any_other_dependency_to_module_than(
        self, dependent: Module, dependent_upon: Module
    ) -> List[Module]:
        # submodules of and including the dependent do not count as allowed imports
        nodes_to_exclude = self._get_all_submodules_of(dependent)

        nodes_fulfilling_criteria = []
        # submodules of the dependent upon do not count as allowed imports if they should import each other
        nodes_that_count_as_not_fulfilling_criterion = self._get_all_submodules_of(
            dependent_upon
        )

        if dependent_upon.parent_module is not None:
            # if the dependent upon module is defined via a parent module, this is not included in the list of modules
            # that do not fulfill the criteria - only its true submodules
            nodes_that_count_as_not_fulfilling_criterion.remove(
                dependent_upon.parent_module
            )

        nodes_to_check = list(nodes_that_count_as_not_fulfilling_criterion)

        if dependent.parent_module is not None:
            # if the dependent module is defined via a parent module, this parent module counts as an allowed import
            nodes_to_exclude.remove(dependent.parent_module)

        checked_nodes = set()

        while nodes_to_check:
            node = nodes_to_check.pop()

            if node in checked_nodes:
                continue

            checked_nodes.add(node)

            parents = self._graph.direct_predecessor_nodes(node)

            for parent in parents:

                if not self._graph.parent_child_relationship(parent, node):
                    if (
                        parent not in nodes_to_exclude
                        and parent not in nodes_that_count_as_not_fulfilling_criterion
                    ):
                        nodes_fulfilling_criteria.append(parent)

                    if parent not in nodes_to_exclude:
                        nodes_to_check.append(parent)

        return self._to_modules(nodes_fulfilling_criteria)

    def _to_modules(self, nodes: List[Node]) -> List[Module]:
        return list(map(lambda node: Module(name=node), nodes))

    def _get_all_submodules_of(self, module: Module) -> Set[Node]:
        """Returns all submodules of a given module.

        Args:
            module: module to retrieve submodules of

        Returns:
            all submodules, including the module itself
        """
        start_node = self._get_node(module)

        nodes_to_check = [start_node]
        checked_nodes = set()

        submodules = set()

        while nodes_to_check:
            node = nodes_to_check.pop()

            if node in checked_nodes:
                continue

            checked_nodes.add(node)
            submodules.add(node)

            children = self._graph.direct_successor_nodes(node)

            for child in children:
                if self._graph.parent_child_relationship(node, child):
                    nodes_to_check.append(child)

        return submodules

    def visualize(self, **kwargs: Any) -> None:
        self._graph.draw(**kwargs)

    def _get_excluded_nodes(self, modules: List[Module]) -> List[Node]:
        result = [module.parent_module for module in modules]
        return [node for node in result if node is not None]



from pathlib import Path
from typing import List

from pytestarch.importer.import_types import Import


class ImportFilter:
    """Filters out imports of external modules from the list of all imports.
    External modules are all modules that are not submodules of the configured module to search for imports.
    """

    def __init__(
        self, exclude_external_libraries: bool, root_module_path: Path
    ) -> None:
        """
        Args:
            exclude_external_libraries: If True, external modules will be filtered out, otherwise this class implements a no-op.
            root_module_path: path to the module that determines which modules are considered external. If a module
                is a submodule of this module, it is considered internal.
        """
        self._exclude_external_libraries = exclude_external_libraries
        self._root_module_path = root_module_path

    def filter(self, imports: List[Import]) -> List[Import]:
        """According to the configuration, imports will be filtered.

        Args:
            imports: list of imports to be filtered
        Returns:
            filtered list of imports
        """
        if not self._exclude_external_libraries:
            return imports

        return [i for i in imports if self._is_internal_import(i)]

    def _is_internal_import(self, i: Import) -> bool:
        importee = i.importee()

        top_level_module = self._root_module_path.name

        return importee.startswith(top_level_module)



import ast
import os
from pathlib import Path
from typing import Optional, Tuple, List

from pytestarch.importer.file_filter import FileFilter
from pytestarch.importer.import_types import NamedModule

PYTHON_FILE_SUFFIX = ".py"


class Parser:
    """Parses all files that match given criteria starting at a source path."""

    def __init__(
        self,
        source_root: Path,
        filter: FileFilter,
        root_and_module_path_identical: bool,
    ) -> None:
        self._source_root = source_root
        self._filter = filter

        self._path_prefix = str(source_root.name)
        self._root_path_is_root_module_path = root_and_module_path_identical

    def parse(self, path: Path) -> Tuple[List[str], List[NamedModule]]:
        """Reads all python files in the given path and returns list of ast
        modules with names.

        Args:
            path: either to a file or to a directory
        Returns:
            list of python modules, one per python file
        """
        self._all_modules = []

        self._root_path_is_root_module_path = path == self._source_root

        paths = [path]
        modules = []

        while paths:
            path = paths.pop()

            if path.is_dir():
                if not self._filter.is_excluded(path):
                    module_name = self._get_module_name(path)
                    if module_name:
                        self._all_modules.append(module_name)
                paths.extend(path.iterdir())
            else:
                new_module = self._parse_file(path)
                if new_module:
                    modules.append(new_module)

        return self._all_modules, modules

    def _parse_file(self, path: Path) -> Optional[NamedModule]:
        """Converts a given python file to an ast module and its name."""
        absolute_path = path.resolve()
        if self._file_should_be_parsed(absolute_path):
            with open(absolute_path, "r") as file:
                code = file.read()

            module_name = self._get_module_name(path)
            self._all_modules.append(module_name)

            return NamedModule(
                ast.parse(code),
                module_name,
            )

    def _get_module_name(self, path: Path) -> str:
        """Determine full name of module, such as A.B.C"""
        module_path = path.relative_to(self._source_root)

        if self._root_path_is_root_module_path:
            if str(module_path) == ".":
                module_path = Path(self._path_prefix)
            else:
                module_path = self._path_prefix / module_path

        module_without_file_suffix = module_path.with_suffix("")
        return str(module_without_file_suffix).replace(os.sep, ".")

    def _file_should_be_parsed(self, path: Path) -> bool:
        """Returns True if path represents a python file that does not match any exclusion filters."""
        correct_file_type = path.suffix == PYTHON_FILE_SUFFIX
        file_excluded = self._filter.is_excluded(path)

        return correct_file_type and not file_excluded



"""Encapsulation of networkx graph functionality."""
from typing import List, Any, Optional

import networkx as nx
from networkx import has_path, draw_networkx, spring_layout

from pytestarch.importer.import_types import Import

EXPECTED_EDGE_AND_NODE_TYPES = "Only str and tuple of two str supported."


Node = str


class Graph:
    """Constructs eval_structure from list of imports.

    Each module passed to this object will be added as a node.
    Importing and imported module are connected via a direct edge.
    Edges are added between each successive level in the module hierarchy.

    E.g.
    Import('A.B', 'C.D.E') results in
    - nodes: ['A', 'A.B', 'C', 'C.D', 'C.D.E']
    - edges: [('A', 'A.B'), ('C', 'C.D'), ('C.D', 'C.D.E'), ('A.B', 'C.D.E')]

    """

    def __init__(
        self,
        all_modules: List[Node],
        imports: List[Import],
        level_limit: Optional[int] = None,
    ) -> None:
        """
        Args:
            all_modules: list of all nodes in the graph, which can be connected by imports.
            imports: all dependencies between the graph's nodes.
            level_limit: if not None, specifies the depth of the graph
        """
        self._all_modules = all_modules
        self._imports = imports
        self._graph = nx.DiGraph()

        self._level_limit = level_limit

        self._initialise()

    def _initialise(self) -> None:
        """Constructs a graph from all modules and their imports."""
        self._add_all_modules_as_nodes()

        for imp in self._imports:
            importer = imp.importer()
            importee = imp.importee()

            self._create_edge(importer, importee)

            self._add_edges_within_module_hierarchy(
                imp.importer_parent_modules(),
                importer,
            )

            all_importee_modules = imp.importee_parent_modules() + [importee]

            for parent, child in zip(
                all_importee_modules[:-1], all_importee_modules[1:]
            ):
                self._create_edge(parent, child, inherits=True)

    def _add_all_modules_as_nodes(self) -> None:
        for module in self._all_modules:
            self._create_node(module)

    def _add_edges_within_module_hierarchy(
        self,
        parent_modules: List[Node],
        child: Node,
    ) -> None:
        """Create edges between a node and its parent recursively until the parent-less parent is reached.

        Args:
            parent_modules: list of all parent modules in order, last element is the direct parent of the child module
            child: lowest element in the module hierarchy
        """
        all_modules = parent_modules + [child]
        for parent, child in zip(all_modules[:-1], all_modules[1:]):
            self._create_edge(parent, child, inherits=True)

    def _create_node(self, node: Node) -> None:
        """Create a node in the dependency graph.
        Args:
            node: to add to the graph
        """
        node = self._flatten_graph_node(node)

        if node not in self._graph:
            self._graph.add_node(node)

    def _create_edge(
        self, node_start: Node, node_end: Node, inherits: bool = False
    ) -> None:
        """Creates an edge in the graph between the two given modules.

        Args:
            node_start: node the edge starts from
            node_end: node the edge points towards
            inherits: if True, edge will be marked as belonging to two nodes that are connected in a parent-child-relationship
        """
        node_start = self._flatten_graph_node(node_start)
        node_end = self._flatten_graph_node(node_end)

        if node_start == node_end:
            return

        # "from foo import bar" - bar could be a function/class in foo
        # or a submodule of foo
        # in order to not add a function/class to the eval_structure, an ende to
        # foo.bar will only be added if both importer
        # and importee are already part of the eval_structure (which they will
        # be if they correspond to modules in the file system
        if (
            self._graph.has_node(node_start)
            and self._graph.has_node(node_end)
            and not self._graph.has_edge(node_start, node_end)
        ):
            self._graph.add_edge(node_start, node_end, inherits=inherits)

    def __contains__(self, item) -> bool:
        if not isinstance(item, (str, tuple)):
            raise TypeError(EXPECTED_EDGE_AND_NODE_TYPES)

        if isinstance(item, str):
            return self._graph.has_node(item)

        if (
            len(item) > 2
            or not isinstance(item[0], str)
            or not isinstance(item[1], str)
        ):
            TypeError(EXPECTED_EDGE_AND_NODE_TYPES)

        return self._graph.has_edge(*item) or has_path(self._graph, *item)

    @property
    def edges_number(self) -> int:
        return self._graph.number_of_edges()

    @property
    def nodes_number(self) -> int:
        return self._graph.number_of_nodes()

    @property
    def nodes(self) -> List[Node]:
        return list(self._graph.nodes)

    def direct_predecessor_nodes(self, node: Node) -> List[Node]:
        """Returns all nodes that have a directed edge towards the given node.

        Args:
            node: node for which to retrieve predecessor nodes

        Returns:
            all predecessor nodes
        """
        return sorted(self._graph.predecessors(node))

    def direct_successor_nodes(self, node: Node) -> List[Node]:
        """Returns all nodes that the given node has a directed edge towards.

        Args:
            node: node for which to retrieve successor nodes

        Returns:
            all successor nodes
        """
        return sorted(self._graph.successors(node))

    def parent_child_relationship(
        self, supposed_parent_node: Node, supposed_child_node: Node
    ) -> bool:
        """Returns True if the given nodes are marked as a parent-child hierarchy.

        Args:
            supposed_parent_node:
            supposed_child_node:

        Returns:
            True if supposed parent is actually parent of supposed child node
        """
        edge_data = self._graph.get_edge_data(supposed_parent_node, supposed_child_node)

        return edge_data["inherits"]

    def draw(self, **kwargs: Any) -> None:
        """Creates a matplotlib plot representing the graph."""
        if "spacing" in kwargs:
            spacing = kwargs.pop("spacing")
            pos = spring_layout(self._graph, k=spacing, iterations=20)
            kwargs["pos"] = pos

        draw_networkx(self._graph, **kwargs)

    def _flatten_graph_node(self, node: Node) -> Node:
        """Limits the depth of the graph by aggregating sub modules above a certain limit.
        Args:
            node: name of node to flatten.

        Returns: flattened name, i.e. only containing at most level limit "." module separators
        """
        if self._level_limit is None:
            return node

        # level 1: include all children of base module:
        # base module: src
        # children: src.A, src.B
        # not to be included: src.A.a, src.A.aa, src.B.b, ...
        # logic: include at most level "." in the node name

        node_parts = node.split(".")
        return ".".join(node_parts[: self._level_limit + 1])



from dataclasses import dataclass
from typing import Tuple


@dataclass
class Config:
    """Configuration of the file parser.

    Attributes:
        excluded_directories: Directory paths not to include in parsing. Can be used to exclude e.g. test directories.
    """

    excluded_directories: Tuple[str, ...]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pytestarch',
 'pytestarch.config',
 'pytestarch.eval_structure',
 'pytestarch.importer',
 'pytestarch.query_language']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5.3,<4.0.0', 'networkx>=2.8.6,<3.0.0']

setup_kwargs = {
    'name': 'pytestarch',
    'version': '1.0.1',
    'description': 'Test framework for software architecture based on imports between modules',
    'long_description': '# Welcome to PyTestArch\n\nPyTestArch is an open source library that allows users to define architectural rules and test their code against them. It is \ngenerally inspired by [ArchUnit](https://www.archunit.org/).\n\n## Installation Guide\nPyTestArch is available via [PyPI](https://pypi.org/project/pytestarch/) and can be installed e.g. via pip: `pip install pytestarch`.\n\n## Usage Guide\nThree steps are required to test an architectural rule:\n\n1) Create an evaluable representation of the source code you want to test\n\n```\nfrom pytestarch.pytestarch import get_evaluable_architecture\n\nevaluable = get_evaluable_architecture("/home/dummy/project", "/home/dummy/project/src")\n```\nThis will scan all python files under /home/dummy/project/src for imports and build an internal representation that can\nlater be queried. The first parameter /home/dummy/project helps PyTestArch to differentiate between internal and external \ndependencies. This evaluable can be used for multiple architectural rule checks; if you are using [pytest](https://docs.pytest.org/en/7.1.x/),\nyou could use a fixture for this evaluable object.\n\n2) Define an architectural rule\n```\nfrom pytestarch.query_language.base_language import Rule\n\nrule = Rule()\n        .modules_that()\n        .are_named("src.moduleB")\n        .should_not()\n        .be_imported_by_modules_that()\n        .are_sub_modules_of("src.moduleA"),\n```\n\nThis rule represents the architectural requirements that a module named "src.moduleB" should not be imported by any module\nthat is a submodule of "src.moduleA", excluding "src.moduleA" itself.\n\n3) Evaluate your code against this rule\n\n```\nrule.assert_applies(evaluable)\n```\nThat\'s it!\n',
    'author': 'zyskarch',
    'author_email': 'zyskarch@gmail.com',
    'maintainer': 'zyskarch',
    'maintainer_email': 'zyskarch@gmail.com',
    'url': 'https://github.com/zyskarch/pytestarch',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Abstract types to specify the interface of evaluable objects."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol, Optional, Any, Tuple, List

from pytestarch.eval_structure.graph import Node


class EvaluableArchitecture(Protocol):
    def get_dependency(
        self, dependent: Module, dependent_upon: Module
    ) -> Optional[Tuple[str, str]]:
        """Returns True if the dependent module is indeed
        depending on the dependent_upon module.
        Submodules of dependent towards are taken into account, but submodules of dependent_upon are not.
        If one or both of the modules are defined by their parent module, this parent module is excluded from possible
        matches.

        Args:
            dependent: Module
            dependent_upon: Module

        Returns:
            Importer and importee if there are any that are sub modules of dependent and dependent_upon respectively.
        """
        raise NotImplementedError()

    def any_dependency_to_module_other_than(
        self, dependent: Module, dependent_upon: Module
    ) -> List[Module]:
        """Returns True if the dependent module has any
        dependency to a module other than the dependent_upon module
        or any of its submodules.
        If the dependent module is defined via a parent module, this parent module is not taken into account.
        If the dependent upon module is defined via a parent module, this parent module counts as an 'other' dependency.

        Args:
            dependent: Module
            dependent_upon: Module

        Returns:
            All modules other than dependent_upon on which dependent module as any dependency
        """
        raise NotImplementedError()

    def any_other_dependency_to_module_than(
        self, dependent: Module, dependent_upon: Module
    ) -> List[Module]:
        """Returns True if any module other than the dependent
        module and its submodules has any dependency to the
        dependent_upon module.
        If the dependent module is defined via a parent module, this parent module is taken into account. This means
        that if the dependent module's parent module has a dependency to the dependent upon module, this will be contained
        in the returned list.
        If the dependent upon module is defined via a parent module, this parent module is not taken into account.

        Args:
            dependent: Module
            dependent_upon: Module

        Returns:
            All modules other than dependent that have any dependency on the dependent upon module
        """
        raise NotImplementedError()

    def visualize(self, **kwargs: Any) -> None:
        """Uses matplotlib to draw the underlying dependency structure.

        Args:
            **kwargs: Any formatting options available for networkx' drawing function, as this is currently the only
                available backend.
                Exception: If 'spacing' is set, this will be interpreted as the parameter 'k' of the spring layout (https://networkx.org/documentation/stable/reference/generated/networkx.drawing.layout.spring_layout.html#networkx.drawing.layout.spring_layout).
        """
        raise NotImplementedError()


@dataclass(frozen=True)
class Module:
    """Represents a python module.
    The module can either be identified by its name or by the name of a parent module.

    Attributes:
        name: full name of the module
        parent_module: full name of the parent module
    """

    name: Optional[str] = None
    parent_module: Optional[Node] = None



from __future__ import annotations

from abc import abstractmethod, ABC
from collections import defaultdict
from typing import Optional, Type, Generic, TypeVar

from pytestarch.eval_structure.eval_structure_types import EvaluableArchitecture, Module
from pytestarch.eval_structure.graph import Node
from pytestarch.query_language.exceptions import ImproperlyConfigured
from pytestarch.query_language.rule_matcher import (
    RuleMatcher,
    DefaultRuleMatcher,
    BehaviorRequirement,
    ModuleRequirement,
    RuleViolations,
)

# (Import, negated, singular)
PREFIX_MAPPING = defaultdict(str)
PREFIX_MAPPING.update(
    {
        (False, False, True): "is ",
        (False, True, True): "is not ",
        (True, True, True): "does not ",
        (False, False, False): "are ",
        (False, True, False): "are not ",
        (True, True, False): "do not ",
    }
)


class RuleApplier(ABC):
    @abstractmethod
    def assert_applies(self, evaluable: EvaluableArchitecture) -> None:
        """
        Calculates whether it (the rule) applies to a given EvaluableArchitecture.
        This means calculating which behavior is wanted and then
        checking this to the state the Evaluable represents.

        Args:
            evaluable: module dependency structure to compare the rule against

        Raises:
            AssertionError: if the rule does not apply to the evaluable object
        """
        raise NotImplementedError()


class BehaviorSpecification(ABC):
    """Offers functionality to specify whether dependencies are expected or not."""

    @abstractmethod
    def should(self) -> DependencySpecification:
        pass

    @abstractmethod
    def should_only(self) -> DependencySpecification:
        pass

    @abstractmethod
    def should_not(self) -> DependencySpecification:
        pass


ModuleSpecificationSuccessor = TypeVar(
    "ModuleSpecificationSuccessor", BehaviorRequirement, RuleApplier
)


class ModuleSpecification(Generic[ModuleSpecificationSuccessor], ABC):
    """Offers functionality to specify detail information about Rule Subjects or Objects."""

    @abstractmethod
    def are_sub_modules_of(self, modules: Node) -> ModuleSpecificationSuccessor:
        pass

    @abstractmethod
    def are_named(self, name: str) -> ModuleSpecificationSuccessor:
        pass


class RuleObject(ModuleSpecification[RuleApplier], ABC):
    pass


class RuleSubject(ModuleSpecification[BehaviorSpecification], ABC):
    pass


class DependencySpecification(ABC):
    """Offers functionality to specify which kind of dependencies are expected."""

    @abstractmethod
    def import_modules_that(self) -> RuleObject:
        pass

    @abstractmethod
    def be_imported_by_modules_that(self) -> RuleObject:
        pass

    @abstractmethod
    def import_modules_except_modules_that(self) -> RuleObject:
        pass

    @abstractmethod
    def be_imported_by_modules_except_modules_that(self) -> RuleObject:
        pass


class RuleBase(ABC):
    """Entry point to each architectural rule."""

    @abstractmethod
    def modules_that(self) -> RuleSubject:
        pass


class Rule(
    DependencySpecification,
    RuleBase,
    BehaviorSpecification,
    RuleObject,
    RuleSubject,
    RuleApplier,
):
    """Represents an architectural rule of the form
    Module1 [verb, such as 'should'] [import type, such as 'import'] Module2
    """

    def __init__(
        self, rule_matcher_class: Type[RuleMatcher] = DefaultRuleMatcher
    ) -> None:
        self._rule_matcher_class = rule_matcher_class

        self._module_to_check: Optional[Module] = None
        self._module_to_check_against: Optional[Module] = None

        self._should = False
        self._should_only = False
        self._should_not = False

        self._except_present = False

        self._import = None

        self._modules_to_check_to_be_specified_next = None

    @property
    def rule_subject(self) -> Optional[Module]:
        return self._module_to_check

    def modules_that(self) -> RuleSubject:
        self._modules_to_check_to_be_specified_next = True
        return self

    def are_sub_modules_of(self, module: Node) -> BehaviorSpecification:
        if self._modules_to_check_to_be_specified_next is None:
            raise ImproperlyConfigured("Specify a RuleSubject or RuleObject first.")

        module = Module(parent_module=module)

        if self._modules_to_check_to_be_specified_next:
            self._module_to_check = module
        else:
            self._module_to_check_against = module

        return self

    def are_named(self, name: str) -> BehaviorSpecification:
        if self._modules_to_check_to_be_specified_next is None:
            raise ImproperlyConfigured("Specify a RuleSubject or RuleObject first.")

        module = Module(name=name)

        if self._modules_to_check_to_be_specified_next:
            self._module_to_check = module
        else:
            self._module_to_check_against = module

        return self

    def should(self) -> DependencySpecification:
        self._should = True
        return self

    def should_only(self) -> DependencySpecification:
        self._should_only = True
        return self

    def should_not(self) -> DependencySpecification:
        self._should_not = True
        return self

    def import_modules_that(self) -> RuleObject:
        self._import = True
        self._modules_to_check_to_be_specified_next = False
        return self

    def be_imported_by_modules_that(self) -> RuleObject:
        self._import = False
        self._modules_to_check_to_be_specified_next = False
        return self

    def import_modules_except_modules_that(self) -> RuleObject:
        self._import = True
        self._except_present = True
        self._modules_to_check_to_be_specified_next = False
        return self

    def be_imported_by_modules_except_modules_that(self) -> RuleObject:
        self._import = False
        self._except_present = True
        self._modules_to_check_to_be_specified_next = False
        return self

    def assert_applies(self, evaluable: EvaluableArchitecture) -> None:
        self._assert_required_configuration_present()

        module_requirement = ModuleRequirement(
            self._module_to_check, self._module_to_check_against, self._import
        )
        behavior_requirement = BehaviorRequirement(
            self._should,
            self._should_only,
            self._should_not,
            self._except_present,
        )

        rule_violations = self._rule_matcher_class(
            module_requirement, behavior_requirement
        ).match(evaluable)

        if rule_violations:
            raise AssertionError(self._create_rule_violation_message(rule_violations))

    def __str__(self) -> str:
        method_name = f'{"should" if self._should else "should only" if self._should_only else "should not"}'

        return (
            f"{self._add_rule_subject_plural_marker_if_needed(self._get_module_name(self._module_to_check))} "
            f"{method_name} "
            f'{"import" if self._import else "be imported by"} '
            f'{"modules except " if self._except_present else ""}'
            f'modules that are {"named" if self._module_to_check_against.name is not None else "sub modules of"} '
            f'"{self._get_module_name(self._module_to_check_against)}".'
        )

    def _get_module_name(self, module: Module) -> str:
        return module.name if module.name is not None else module.parent_module

    def _assert_required_configuration_present(self) -> None:
        behavior_missing = not any([self._should, self._should_only, self._should_not])
        dependency_missing = self._import is None
        subject_missing = not self._module_to_check
        object_missing = not self._module_to_check_against

        if any([behavior_missing, dependency_missing, subject_missing, object_missing]):
            subject_message = self._name_or_empty(subject_missing, RuleSubject)
            behavior_message = self._name_or_empty(
                behavior_missing, BehaviorSpecification
            )
            dependency_message = self._name_or_empty(
                dependency_missing, DependencySpecification
            )
            object_message = self._name_or_empty(object_missing, RuleObject)

            messages = [
                subject_message,
                behavior_message,
                dependency_message,
                object_message,
            ]

            error_message = (
                f"Specify {', '.join(filter(lambda m: len(m) > 0, messages))}."
            )

            raise ImproperlyConfigured(error_message)

    @classmethod
    def _name_or_empty(cls, empty: bool, clz: type) -> str:
        return f"a {clz.__name__}" if empty else ""

    def _create_rule_violation_message(self, rule_violations: RuleViolations) -> str:
        rule_subject = self._create_rule_subject_for_rule_violation_message(
            rule_violations
        )
        verb = self._create_verb_for_rule_violation_message(rule_violations)
        rule_object = self._create_rule_object_for_rule_violation_message(
            rule_violations
        )
        return f"{rule_subject} {verb}{rule_object}."

    def _create_rule_subject_for_rule_violation_message(
        self, rule_violations: RuleViolations
    ) -> str:
        module_name = self._get_module_name(self._module_to_check)

        if self._forbidden_import(rule_violations):
            if rule_violations.strict_dependency is not None:
                module_name = self._get_subject_of_strict_dependency(rule_violations)

                # no plural marker needed, as module is always singular
                return f'"{module_name}"'

        return self._add_rule_subject_plural_marker_if_needed(module_name)

    def _get_subject_of_strict_dependency(self, rule_violations: RuleViolations) -> str:
        # strict dependency is always reported in format (importer, importee)
        # if rule is of format A ... imports B, the dependency will be (A, B)
        idx = int(not self._import)
        return rule_violations.strict_dependency[idx]

    def _get_object_of_strict_dependency(self, rule_violations: RuleViolations) -> str:
        # strict dependency is always reported in format (importer, importee)
        # if rule is of format A ... be imported by B, the dependency will be (B, A)
        # but in our rule, B is the object
        idx = int(self._import)
        return rule_violations.strict_dependency[idx]

    def _add_rule_subject_plural_marker_if_needed(self, module_name: str) -> str:
        prefix = ""
        if not self._rule_subject_singular:
            prefix = "Sub modules of "
        return f'{prefix}"{module_name}"'

    def _create_verb_for_rule_violation_message(
        self, rule_violations: RuleViolations
    ) -> str:
        base_verb = "import" if self._import else "imported by"

        suffix = " "

        if rule_violations.should_violated:
            negated = True

        elif (
            rule_violations.should_except_violated
            or rule_violations.should_only_except_violated_by_no_import
        ):
            negated = True

        elif (
            rule_violations.should_only_except_violated_by_forbidden_import
            or rule_violations.should_not_violated
            or rule_violations.should_only_violated
            or rule_violations.should_not_except_violated
        ):

            if self._import and (
                self._rule_subject_singular or (self._forbidden_import(rule_violations))
            ):
                suffix = "s "

            negated = False
        else:
            raise Exception("Unknown rule violation detected.")

        prefix = PREFIX_MAPPING[
            (
                self._import,
                negated,
                self._rule_subject_singular
                or (self._forbidden_import(rule_violations)),
            )
        ]

        return f"{prefix}{base_verb}{suffix}"

    def _forbidden_import(self, rule_violations: RuleViolations) -> bool:
        return (
            rule_violations.should_not_violated
            or rule_violations.should_only_except_violated_by_forbidden_import
        )

    @property
    def _rule_subject_singular(self) -> bool:
        return self._module_to_check.name is not None

    @property
    def _rule_object_singular(self) -> bool:
        return self._module_to_check_against.name is not None

    def _create_rule_object_for_rule_violation_message(
        self, rule_violations: RuleViolations
    ) -> str:
        object_name = self._get_module_name(self._module_to_check_against)
        module_qualifier = "" if self._rule_object_singular else "a sub module of "

        if (
            rule_violations.should_except_violated
            or rule_violations.should_only_except_violated_by_no_import
        ):
            module_qualifier = f"any module that is not {'' if self._rule_object_singular else 'a sub module of '}"
        elif self._forbidden_import(rule_violations):
            module_qualifier = ""

            if rule_violations.strict_dependency is not None:
                object_name = self._get_object_of_strict_dependency(rule_violations)

        if (
            rule_violations.should_not_except_violated
            or rule_violations.should_only_violated
        ):
            object_name = ", ".join(
                (module.name for module in rule_violations.lax_dependencies_found)
            )

        return f'{module_qualifier}"{object_name}"'



from pathlib import Path

from pytestarch.config.config import Config

ALL_MARKER = "*"


class FileFilter:
    """Uses a pseudo-regex pattern to determine whether a file or directory should be excluded.

    Allowed patterns are:
    - *A*: file or dir will be excluded, if 'A' is part of their name
    - *A: file or dir will be excluded, if their name ends with 'A'
    - A*: file or dir will be excluded, if their name starts with 'A'
    - A: file or dir will be excluded, if their name is 'A'
    """

    def __init__(self, config: Config) -> None:
        self._excluded_dirs = set()
        self._excluded_prefixes = set()
        self._excluded_suffixes = set()
        self._excluded_infixes = set()

        for dir in config.excluded_directories:
            if dir.startswith(ALL_MARKER) and dir.endswith(ALL_MARKER):
                self._excluded_infixes.add(dir[1:-1])
            elif dir.startswith(ALL_MARKER):
                self._excluded_suffixes.add(dir[1:])
            elif dir.endswith(ALL_MARKER):
                self._excluded_prefixes.add(dir[:-1])
            else:
                self._excluded_dirs.add(dir)

    def is_excluded(self, path: Path) -> bool:
        """Returns True if the path matches one of the pre-configured exclusion patterns."""
        path = str(path)

        full_excluded = path in self._excluded_dirs
        prefix_excluded = any(
            [True for prefix in self._excluded_prefixes if path.startswith(prefix)]
        )
        suffix_excluded = any(
            [True for suffix in self._excluded_suffixes if path.endswith(suffix)]
        )
        infix_excluded = any(
            [True for infix in self._excluded_infixes if infix in path]
        )

        return full_excluded or prefix_excluded or suffix_excluded or infix_excluded
